/*
 * Segment 2 (KEx),  SWE2.4 Email/Fax/SMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.emailfaxsms.sms.sender;

import java.util.Collection;
import java.util.LinkedList;

import de.bsvrz.kex.emailfaxsms.AsyncSerialPortCommunicator;
import de.bsvrz.kex.emailfaxsms.AsyncSerialPortCommunicator.ModemType;
import de.bsvrz.kex.emailfaxsms.CommunicationException;
import de.bsvrz.kex.emailfaxsms.VersandStatus;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import gnu.io.NoSuchPortException;

/**
 * Versendet SMS Nachrichten im Text Modus.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class KexSmsSender {

	/**
	 * Textmode SMS maximale L&auml;nge.
	 */
	public static final int GSM_TEXTMODE_SMS_LENGTH = 160;

	/**
	 * LOGGER.
	 */
	protected static final Debug LOGGER = Debug.getLogger();

	/**
	 * Konstanten.
	 */
	/** End of file. */
	protected static final byte EOF = 0x1A;
	/** Escape. */
	protected static final byte ESC = 0x1B;

	/**
	 * die Antwort bei SMS-Eingabe im Text mode.
	 */
	protected static final byte[] WRITE_INPUT = "> ".getBytes();

	/**
	 * Antwort auf +CPIN?, wenn SIM Karte aktiviert ist.
	 */
	protected static final String READY = "+CPIN:<SP>READY";

	/**
	 * Antwort auf +CPIN?, wenn SIM Karte gesperrt ist.
	 */
	protected static final String SIM_PIN = "+CPIN:<SP>SIM<SP>PIN";

	/**
	 * ComPort.
	 */
	protected String com;
	/**
	 * Baudrate.
	 */
	protected int baudrate;
	/**
	 * Standardtimeout.
	 */
	protected long timeout = 20000;
	/**
	 * Pin-Kode f&uuml;r die Authentifizierung im GSM-Netz.
	 */
	protected String gsmPin;
	/**
	 * Pr&auml;fix f&uuml;r jede Nummer.
	 */
	protected String prefix = "";
	/**
	 * InitSequenz.
	 */
	protected String initSeq = "AT";
	/**
	 * Modemtyp.
	 */
	protected ModemType modemType = ModemType.MODEM_GSM;

	/**
	 * Standardkonstruktor.
	 *
	 * @param com
	 *            COM-Port
	 * @param baud
	 *            Baudrate
	 * @param timeout
	 *            Timeout in Sekunden
	 * @param gsmPin
	 *            PIN
	 * @param typ
	 *            Modem-Typ
	 * @param initSequenz
	 *            Initstring
	 * @param prefix
	 *            Wahl-Prefix
	 * @throws Exception
	 *             Feher beim Zugriff auf serielle Schnittstelle
	 */
	public KexSmsSender(final String com, final int baud, final long timeout, final String gsmPin, final ModemType typ,
			final String initSequenz, final String prefix) throws Exception {
		this.timeout = timeout * 1000;
		this.gsmPin = gsmPin;
		this.com = com;
		this.baudrate = baud;
		this.gsmPin = gsmPin;
		this.initSeq = initSequenz;
		this.prefix = prefix;
		this.modemType = typ;

		AsyncSerialPortCommunicator comm;
		try {
			comm = new AsyncSerialPortCommunicator(com, baud, 512, 512, 5000, false);
		} catch (final UnsatisfiedLinkError ule) {
			KexSmsSender.LOGGER.error(ule.getMessage());
			throw new NoSuchPortException();

		}
		try {
			comm.sendCmd("AT");
			comm.recvATantwortMitKontroll();
		} catch (final CommunicationException e) {
			KexSmsSender.LOGGER.warning("Keine Antwort des Kommunikationsgerätes an " + com);
			comm.beenden();
			throw e;
		}
		comm.beenden();
		KexSmsSender.LOGGER.fine("SmsSender Initialisierung komplett");
	}

	/**
	 * L&ouml;scht alle gespeicherten SMS.
	 *
	 * @param comm
	 *            ComPort
	 * @param n
	 *            Kapazit&auml;t des Speichers
	 * @throws CommunicationException
	 *             bei Ausnahmen
	 */
	public void deleteMemory(final AsyncSerialPortCommunicator comm, final int n) throws CommunicationException {

		for (int i = 1; i < n; i++) {
			comm.sendCmd("AT+CMGD=" + i);
			comm.recvATantwortMitKontroll();
		}
	}

	/**
	 * Sendet eine SMS.
	 *
	 * @param empfaenger
	 *            Empf&auml;nger
	 * @param text
	 *            Text in Latin
	 * @param debug
	 *            Debug-Flag
	 * @return VersandStatus
	 */
	public VersandStatus send(final String empfaenger, final String text, final boolean debug) {
		AsyncSerialPortCommunicator comm = null;
		synchronized (AsyncSerialPortCommunicator.getLock(com)) {
			try {
				comm = new AsyncSerialPortCommunicator(com, baudrate, 512, 512, timeout, false);
				if (debug) {
					comm.setDebug(debug);
				}
				final Collection<String> subStrings = teileNachricht(text);
				for (final String s : subStrings) {
					sendInTextModus(comm, empfaenger, s, this.gsmPin);
				}
				comm.beenden();
			} catch (final Exception e) {
				final String s = "SMS Nicht versendet: " + e.getMessage();
				KexSmsSender.LOGGER.warning(s);
				MessageSender.getInstance().sendMessage(MessageType.APPLICATION_DOMAIN, MessageGrade.WARNING, s);
				if (comm != null) {
					comm.beenden();
				}
				return new VersandStatus(s, false);
			}
		}
		return VersandStatus.VERSAND_OK;
	}

	/**
	 * Sendet eine SMS Nachricht im Text-Modus.
	 *
	 * @param comm
	 *            ComPort
	 * @param empfaenger
	 *            Empf&auml;nger
	 * @param textGSM
	 *            Text
	 * @param pin
	 *            Pin-Code
	 * @throws CommunicationException
	 *             Fehler bei der Kommunikation
	 */
	public void sendInTextModus(final AsyncSerialPortCommunicator comm, final String empfaenger, final String textGSM,
			final String pin) throws CommunicationException {
		Collection<String> resp;

		comm.sendCmdMitKontroll(initSeq);
		comm.sendCmdMitKontroll("ATE0");

		setCpinReady(comm, pin);

		comm.sendCmdMitKontroll("AT+CSCS=GSM");
		comm.sendCmdMitKontroll("AT+CMGF=1");
		comm.sendCmdMitKontroll("AT+CSMP=17,167,0,0");

		comm.sendCmd("AT+CMGS=" + prefix + empfaenger);
		comm.recvBytes(KexSmsSender.WRITE_INPUT); // Kommt immer, es gibt keine
													// andere
		// Antwort
		comm.sendBytes(textGSM.getBytes());
		comm.sendByte(KexSmsSender.EOF);
		resp = comm.recvATantwortMitKontroll();
		ueberPruefeAntwort(resp, "+CMGS:<SP>", "SMS wurde nicht versendet (Fehlerhafte Eingabe)");
	}

	/**
	 * Entsperrt ein GSM-Faxmodem.
	 *
	 * @param comm
	 *            ComPort
	 * @param pin
	 *            Pin-Code
	 * @throws CommunicationException
	 *             bei Ausnahmen
	 */
	public void setCpinReady(final AsyncSerialPortCommunicator comm, final String pin) throws CommunicationException {
		final String cPIN = "+CPIN";
		boolean cpinUnterstuetzt = false;
		Collection<String> antwort = null;
		try {
			comm.sendCmd("AT+CPIN?");
			antwort = comm.recvATantwortMitKontroll();
		} catch (final CommunicationException e) {
			// Modem unterstuetzt nicht CPIN d.H. es ist kein GSM modem
			if (modemType == ModemType.MODEM_GSM) {
				KexSmsSender.LOGGER.warning("Modemtyp ist GSM, aber das Gerät unterstützt das Kommando +CPIN nicht");
			}
			return;
		}

		for (final String s : antwort) {
			if (s.startsWith(KexSmsSender.READY)) {
				return;
			} else if (s.startsWith(KexSmsSender.SIM_PIN)) {
				try {
					comm.sendCmd("AT+CPIN=" + pin);
					comm.recvATantwortMitKontroll();
					// Nach der entsperrung, wenn man sofort eine SMS versendet
					// bekommt man oft ein ERRORs
					Thread.sleep(10000);
				} catch (final CommunicationException e) {
					throw new CommunicationException("SIM-Karte konnte nicht entsperrt werden");
				} catch (final InterruptedException e) {
					KexSmsSender.LOGGER.warning("Sleep nach PIN-Eingabe wurde unterbrochen");
				}
				return;
			} else if (s.startsWith(cPIN)) {
				cpinUnterstuetzt = true;
			}
		}
		if (!cpinUnterstuetzt) {
			return;
		}

		throw new CommunicationException("SIM-Karte konnte nicht entsperrt werden");
	}

	/**
	 * Teilt die Nachricht so auf, damit die 160 Charakter per SMS optimal
	 * benutzt, aber W&ouml;rter nicht gebrochen werden. Wenn nicht m&ouml;glich
	 * (der Text enth&auml;lt keine Leerzeichen) dann wird die SMS an der
	 * 160-Zeichen-Grenze gebrochen.
	 *
	 * @param text
	 *            Text in GSM Alphabet
	 * @return Liste von Strings, alle k&uuml;rzer als 160 Zeichen
	 */
	protected Collection<String> teileNachricht(final String text) {
		final Collection<String> teile = new LinkedList<String>();
		final char[] str = text.toCharArray();
		int beg = 0, end = KexSmsSender.GSM_TEXTMODE_SMS_LENGTH;

		while (beg < str.length) {
			if (str.length > (end + 1)) {
				int i;
				for (i = 0; i < 10; i++) {
					if (str[end - i] == MapLatin1GSM.GSMSP) {
						break;
					}
				}
				if (i < 10) {
					end -= i;
				}
			} else {
				end = str.length;
			}

			teile.add(new String(str, beg, end - beg));

			beg = end;
			if ((beg < str.length) && (str[beg] == MapLatin1GSM.GSMSP)) {
				beg++;
			}
			end = beg + KexSmsSender.GSM_TEXTMODE_SMS_LENGTH;
		}
		return teile;
	}

	/**
	 * &Uuml;berpr&uuml;ft, ob die Antwort einen gesuchten String enth&auml;lt.
	 *
	 * @param antwort
	 *            Antwort des Faxger&auml;ts
	 * @param gesucht
	 *            Gesuchte Zeichenkette
	 * @param fehlermeldung
	 *            Fehlermeldung, wenn gesuchte Zeichenkette nicht gefunden wurde
	 * @return true, wenn gesuchte Zeichenkette gefunden wurde
	 * @throws CommunicationException
	 *             wird geworfen, wenn gesuchte Zeichenkette nicht gefunden
	 *             wurde
	 */
	public boolean ueberPruefeAntwort(final Collection<String> antwort, final String gesucht, final String fehlermeldung)
			throws CommunicationException {
		for (final String s : antwort) {
			if (s.startsWith(gesucht)) {
				return true;
			}
		}
		throw new CommunicationException(fehlermeldung);
	}
}
